/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server;

import io.vertx.core.AsyncResult;
import io.vertx.core.DeploymentOptions;
import io.vertx.core.Vertx;
import io.vertx.core.json.JsonObject;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import net.as205315.applications.ora.server.lobby.OraLobbyVerticle;
import net.as205315.applications.ora.server.management.ManagementVerticle;
import net.as205315.applications.ora.server.proxy.OraProxyVerticle;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class Main {

    public static void main(String[] args) throws IOException {
        Vertx vertx = Vertx.vertx();

        JsonObject config = new JsonObject();
        putString(config, Config.CONFIG_LOBBY_ADDRESS);
        putInteger(config, Config.CONFIG_LOBBY_PORT);
        putString(config, Config.CONFIG_PROXY_ADDRESS);
        putInteger(config, Config.CONFIG_PROXY_PORT);
        putString(config, Config.CONFIG_MGMT_ADDRESS);
        putInteger(config, Config.CONFIG_MGMT_PORT);
        putString(config, Config.CONFIG_CLIENT_ADDRESS);
        putInteger(config, Config.CONFIG_CLIENT_PORT);
        putString(config, Config.CONFIG_LOBBY_NAME);
        putString(config, Config.CONFIG_LOBBY_GROUP);
        putBoolean(config, Config.CONFIG_LOBBY_ENABLED);
        putBoolean(config, Config.CONFIG_DEV_MODE);

        deployLobby(config, vertx);
    }

    private static void deployLobby(JsonObject config, Vertx vertx) {
        if (config.getBoolean(Config.CONFIG_LOBBY_ENABLED, false)) {
            vertx.deployVerticle(OraLobbyVerticle.class, new DeploymentOptions().setConfig(config), res -> {
                processResult(res, "Deployment of lobby verticle failed.");

                deployProxy(config, vertx);
            });
        } else {
            deployProxy(config, vertx);
        }
    }

    private static void deployProxy(JsonObject config, Vertx vertx) {
        vertx.deployVerticle(OraProxyVerticle.class, new DeploymentOptions().setConfig(config), res -> {
            processResult(res, "Deployment of proxy verticle failed.");

            deployManagement(config, vertx);
        });
    }

    private static void deployManagement(JsonObject config, Vertx vertx) {
        vertx.deployVerticle(ManagementVerticle.class, new DeploymentOptions().setConfig(config), res -> {
            processResult(res, "Deployment of management verticle failed.");

            deployDevelopmentUtils(config, vertx);
        });
    }

    private static void deployDevelopmentUtils(JsonObject config, Vertx vertx) {
        if (config.getBoolean(Config.CONFIG_DEV_MODE, false)) {
            try {
                String lobbyConfig = new BufferedReader(new InputStreamReader(Main.class.getResource("/" + Main.class.getPackage().getName().replace(".", "/") + "/dev.json").openStream())).lines().collect(Collectors.joining("\n"));
                vertx.eventBus().send(OraLobbyVerticle.EVENTBUS_ADDRESS_ADD, config.copy().put(Config.CONFIG_LOBBY_CONFIG, new JsonObject(lobbyConfig)), res -> {
                    processResult(res, "Start of dev lobby failed.");
                });
            } catch (IOException ex) {
                throw new RuntimeException("Failed to load dev config.", ex);
            }
        }
    }

    private static void putString(JsonObject config, String key) {
        Optional.ofNullable(System.getProperty(key, System.getenv(key))).ifPresent(value -> config.put(key, value));
    }

    private static void putInteger(JsonObject config, String key) {
        Optional.ofNullable(System.getProperty(key, System.getenv(key))).ifPresent(value -> config.put(key, Integer.valueOf(value)));
    }

    private static void putBoolean(JsonObject config, String key) {
        Optional.ofNullable(System.getProperty(key, System.getenv(key))).ifPresent(value -> config.put(key, Boolean.valueOf(value)));
    }

    private static Object processResult(AsyncResult<?> result, String msg) {
        if (result.failed()) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, msg, result.cause());
            System.exit(1);
            return null;
        } else {
            return result.result();
        }
    }
}
