/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.authentication;

import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import net.as205315.applications.ora.server.lobby.Client;
import net.as205315.applications.ora.server.packets.AuthenticationErrorOrder;
import net.as205315.applications.ora.server.packets.MessageOrder;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ClientAuthentication {

    private final String remote;
    private final List<Authenticator.Reauthenticator> reauthenticators;
    private Set<Permission> permissions;

    public ClientAuthentication(String remote, List<Authenticator.Reauthenticator> reauthenticators) {
        this.remote = remote;
        this.reauthenticators = reauthenticators;
        updateCache(true);
    }

    public String getRemote() {
        return remote;
    }

    public boolean hasPermission(Permission permission) {
        updateCache(false);
        return permissions.contains(permission);
    }

    public boolean requirePermissionOrDisconnect(Client client, Permission permission) {
        if (hasPermission(permission)) {
            return true;
        } else {
            client.sendPacket(new AuthenticationErrorOrder(0, 0, "Permission " + permission + " was not granted to you. Maybe authentication failure?"));
            client.disconnect();
            return false;
        }
    }

    public boolean requirePermissionOrInform(Client client, Permission permission) {
        if (hasPermission(permission)) {
            return true;
        } else {
            client.sendPacket(new MessageOrder("Permission " + permission + " was not granted to you. Maybe authentication failure?"));
            return false;
        }
    }

    private void updateCache(boolean force) {
        if (force || reauthenticators.stream().anyMatch(Authenticator.Reauthenticator::needsUpdate)) {
            Set<Permission> temp = reauthenticators.stream()
                    .flatMap(Authenticator.Reauthenticator::getPermissions)
                    .collect(Collectors.toSet());
            permissions = temp.isEmpty() ? EnumSet.noneOf(Permission.class) : EnumSet.copyOf(temp);
        }
    }

    public static enum Permission {
        CONNECT, ADMIN
    }
}
