/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.authentication;

import java.util.EnumSet;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Supplier;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.lobby.ClientInfo;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class DynamicAdminAuthenticator implements Authenticator {

    private final Supplier<Stream<ClientInfo>> clients;
    private final AtomicBoolean updateRunning = new AtomicBoolean();

    public DynamicAdminAuthenticator(Supplier<Stream<ClientInfo>> clients) {
        this.clients = clients;
    }

    private <T> T protectRecursion(Supplier<T> task, Supplier<T> fallback) {
        boolean old = updateRunning.getAndSet(true);
        try {
            return old ? fallback.get() : task.get();
        } finally {
            updateRunning.set(old);
        }
    }

    @Override
    public Reauthenticator authenticate(String username, String password, String remote, ClientInfo clientInfo) {
        return new DynamicAdminReauthenticator();
    }

    private class DynamicAdminReauthenticator implements Reauthenticator {

        @Override
        public boolean needsUpdate() {
            return protectRecursion(() -> clients.get().map(ClientInfo::getClientAuthentication).noneMatch(ca -> ca.hasPermission(ClientAuthentication.Permission.ADMIN)), () -> false);
        }

        @Override
        public Stream<ClientAuthentication.Permission> getPermissions() {
            return protectRecursion(
                    () -> clients.get().map(ClientInfo::getClientAuthentication).noneMatch(ca -> ca.hasPermission(ClientAuthentication.Permission.ADMIN))
                    ? EnumSet.of(ClientAuthentication.Permission.ADMIN).stream()
                    : EnumSet.noneOf(ClientAuthentication.Permission.class).stream(),
                    () -> EnumSet.noneOf(ClientAuthentication.Permission.class).stream());
        }
    }
}
