/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.authentication;

import io.vertx.core.Vertx;
import io.vertx.core.impl.ConcurrentHashSet;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.auth.jwt.JWTAuth;
import io.vertx.ext.auth.jwt.JWTAuthOptions;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.EnumSet;
import java.util.Objects;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.lobby.ClientInfo;
import net.as205315.applications.ora.server.lobby.GlobalSettings;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class JwtAuthenticator implements Authenticator {

    private static final Logger LOGGER = Logger.getLogger(JwtAuthenticator.class.getName());

    private static final String CONFIGURATION;
    private static final Set<String> CONSUMED_TOKENS = new ConcurrentHashSet<>();

    static {
        try {
            String jwtFile = Objects.requireNonNull(System.getProperty(Config.CONFIG_JWT_FILE, System.getenv(Config.CONFIG_JWT_FILE)), "env:" + Config.CONFIG_JWT_FILE);
            CONFIGURATION = Files.readString(new File(jwtFile).toPath(), StandardCharsets.UTF_8);
        } catch (IOException | RuntimeException ex) {
            throw new RuntimeException("Failed to load JWT configuration from " + Config.CONFIG_JWT_FILE + "=\"" + System.getenv(Config.CONFIG_JWT_FILE) + "\".", ex);
        }
    }

    private final GlobalSettings globalSettings;
    private final JWTAuth jwt;

    public JwtAuthenticator(GlobalSettings globalSettings, Vertx vertx) {
        this.globalSettings = globalSettings;
        jwt = JWTAuth.create(vertx, new JWTAuthOptions(new JsonObject(CONFIGURATION)));
    }

    @Override
    public Reauthenticator authenticate(String username, String password, String remote, ClientInfo clientInfo) {
        return new JwtReauthenticator(username, password);
    }

    private class JwtReauthenticator implements Reauthenticator {

        private Set<ClientAuthentication.Permission> permissions = EnumSet.noneOf(ClientAuthentication.Permission.class);
        private boolean dirty;

        public JwtReauthenticator(String user, String token) {
            jwt.authenticate(new JsonObject().put("jwt", token), authenticate -> {
                if (authenticate.failed()) {
                    LOGGER.log(Level.WARNING, "JWT authentication failed for user {0}: {1}", new Object[]{user, authenticate.cause().getMessage()});
                } else {
                    String jti = authenticate.result().principal().getString("jti");
                    if (jti != null) {
                        if (CONSUMED_TOKENS.contains(jti)) {
                            LOGGER.log(Level.WARNING, "Detected reused JWT: {0}.", jti);
                            return;
                        }
                        CONSUMED_TOKENS.add(jti);
                    }
                    for (ClientAuthentication.Permission perm : ClientAuthentication.Permission.values()) {
                        authenticate.result().isAuthorized(globalSettings.getMod() + ":" + perm.name().toLowerCase(), authorize -> {
                            if (authorize.failed()) {
                                LOGGER.log(Level.WARNING, "JWT authorization failed for user {0}: {1}", new Object[]{user, authenticate.cause().getMessage()});
                            } else {
                                if (authorize.result()) {
                                    permissions.add(perm);
                                    dirty = true;
                                }
                            }
                        });
                    }
                }
            });
        }

        @Override
        public boolean needsUpdate() {
            return dirty;
        }

        @Override
        public Stream<ClientAuthentication.Permission> getPermissions() {
            return permissions.stream();
        }
    }
}
