/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.authentication;

import java.util.Arrays;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.commands.CommandHandler;
import net.as205315.applications.ora.server.lobby.Client;
import net.as205315.applications.ora.server.lobby.ClientInfo;
import net.as205315.applications.ora.server.packets.MessageOrder;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class TemporaryGrantAuthenticator implements Authenticator, CommandHandler {

    private final Supplier<Stream<ClientInfo>> clients;
    private final Map<ClientInfo, TemporaryGrantReauthenticator> grants = new WeakHashMap<>();

    public TemporaryGrantAuthenticator(Supplier<Stream<ClientInfo>> clients) {
        this.clients = clients;
    }

    @Override
    public Reauthenticator authenticate(String username, String password, String remote, ClientInfo clientInfo) {
        TemporaryGrantReauthenticator grant = new TemporaryGrantReauthenticator();
        grants.put(clientInfo, grant);
        return grant;
    }

    @Override
    public boolean runCommand(Client client, String command, List<String> args) {
        switch (command) {
            case "auth":
                if (args.size() == 2) {
                    if (client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        AtomicBoolean permissionsOk = new AtomicBoolean(true);
                        Set<ClientAuthentication.Permission> permissions = args.get(1).isEmpty() ? Collections.emptySet() : Arrays.asList(args.get(1).split(",")).stream()
                                .map(value -> {
                                    try {
                                        return ClientAuthentication.Permission.valueOf(value.trim());
                                    } catch (IllegalArgumentException ex) {
                                        permissionsOk.set(false);
                                        client.sendPacket(new MessageOrder("Permission \"" + value + "\" doesn't exist."));
                                        return null;
                                    }
                                }).collect(Collectors.toSet());

                        if (!permissionsOk.get()) {
                            return true;
                        }

                        AtomicBoolean found = new AtomicBoolean();
                        clients.get()
                                .filter(c -> Objects.equals(args.get(0), c.getName()))
                                .map(grants::get).forEach(g -> {
                            found.set(true);
                            g.setPermissions(permissions);
                        });

                        // Trigger refresh early!
                        clients.get().forEach(c -> c.getClientAuthentication().hasPermission(ClientAuthentication.Permission.CONNECT));

                        if (!found.get()) {
                            client.sendPacket(new MessageOrder("Unknown user \"" + args.get(0) + "\""));
                        }
                    }

                    return true;
                } else {
                    return false;
                }
            default:
                return false;
        }
    }

    private class TemporaryGrantReauthenticator implements Reauthenticator {

        private final Set<ClientAuthentication.Permission> permissions = EnumSet.noneOf(ClientAuthentication.Permission.class);
        private boolean updatePending;

        @Override
        public boolean needsUpdate() {
            return updatePending;
        }

        @Override
        public Stream<ClientAuthentication.Permission> getPermissions() {
            updatePending = false;
            return permissions.stream();
        }

        public void setPermissions(Set<ClientAuthentication.Permission> permissions) {
            this.permissions.clear();
            this.permissions.addAll(permissions);
            updatePending = true;
        }
    }
}
