/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.json.JsonObject;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Queue;
import net.as205315.applications.ora.server.authentication.ClientAuthentication;
import net.as205315.applications.ora.server.utils.StringUtils;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ClientInfo {

    private int id;
    private ClientAuthentication clientAuthentication = new ClientAuthentication(null, Collections.emptyList());
    private String preferredColor;
    private String color;
    private String faction;
    private int spawnPoint;
    private String name;
    private String address;
    private ClientState state;
    private int team;
    private String slot;
    private String bot;
    private int botController;
    private final Queue<Long> latencyHistory = new LinkedList<>();

    public ClientInfo(JsonObject obj) {
        this.id = Integer.parseInt(obj.getString("Index"));
        this.preferredColor = StringUtils.emptyToNull(obj.getString("PreferredColor"));
        this.color = StringUtils.emptyToNull(obj.getString("Color"));
        this.faction = StringUtils.emptyToNull(obj.getString("Faction"));
        this.spawnPoint = Integer.parseInt(obj.getString("SpawnPoint"));
        this.name = StringUtils.nullToEmpty(obj.getString("Name"));
        this.address = StringUtils.emptyToNull(obj.getString("Address"));
        this.state = ClientState.parseCsName(obj.getString("State", "Invalid"));
        this.team = Integer.parseInt(obj.getString("Team"));
        this.slot = StringUtils.emptyToNull(obj.getString("Slot"));
        this.bot = StringUtils.emptyToNull(obj.getString("Bot"));
        this.botController = Integer.parseInt(obj.getString("BotControllerClientIndex"));
    }

    public ClientInfo(int id) {
        this.id = id;
    }

    public int getId() {
        return id;
    }

    public void setId(int id) {
        this.id = id;
    }

    public ClientAuthentication getClientAuthentication() {
        return clientAuthentication;
    }

    public void setClientAuthentication(ClientAuthentication clientAuthentication) {
        this.clientAuthentication = clientAuthentication;
    }

    public String getPreferredColor() {
        return preferredColor;
    }

    public void setPreferredColor(String preferredColor) {
        this.preferredColor = preferredColor;
    }

    public String getColor() {
        return color;
    }

    public void setColor(String color) {
        this.color = color;
    }

    public String getFaction() {
        return faction;
    }

    public void setFaction(String faction) {
        this.faction = faction;
    }

    public int getSpawnPoint() {
        return spawnPoint;
    }

    public void setSpawnPoint(int spawnPoint) {
        this.spawnPoint = spawnPoint;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getAddress() {
        return address;
    }

    public void setAddress(String address) {
        this.address = address;
    }

    public ClientState getState() {
        return state;
    }

    public void setState(ClientState state) {
        this.state = state;
    }

    public int getTeam() {
        return team;
    }

    public void setTeam(int team) {
        this.team = team;
    }

    public String getSlot() {
        return slot;
    }

    public void setSlot(String slot) {
        this.slot = slot;
    }

    public String getBot() {
        return bot;
    }

    public void setBot(String bot) {
        this.bot = bot;
    }

    public int getBotController() {
        return botController;
    }

    public void setBotController(int botController) {
        this.botController = botController;
    }

    public void appendLatency(long latencyInNs) {
        latencyHistory.add(latencyInNs);
        int size = latencyHistory.size();
        for (int i = 0; i < (size - 5); i++) {
            latencyHistory.remove();
        }
    }

    public List<Long> getLatencyHistory() {
        return Collections.unmodifiableList(new ArrayList<>(latencyHistory));
    }

    public JsonObject serialize() {
        JsonObject obj = new JsonObject();
        obj.put("Index", id);
        obj.put("PreferredColor", StringUtils.nullToEmpty(preferredColor));
        obj.put("Color", StringUtils.nullToEmpty(color));
        obj.put("Faction", StringUtils.nullToEmpty(faction));
        obj.put("SpawnPoint", spawnPoint);
        obj.put("Name", name);
        obj.put("IpAddress", StringUtils.nullToEmpty(address));
        obj.put("State", state.getCsName());
        obj.put("Team", team);
        obj.put("Slot", StringUtils.nullToEmpty(slot));
        obj.put("Bot", StringUtils.nullToEmpty(bot));
        obj.put("BotControllerClientIndex", botController);
        obj.put("IsAdmin", clientAuthentication.hasPermission(ClientAuthentication.Permission.ADMIN) ? "True" : "False");
        obj.put("IsSpectator", slot == null ? "True" : "False");
        obj.put("IsBot", bot != null ? "True" : "False");
        return obj;
    }

    public enum ClientState {
        NOT_READY("NotReady"), INVALID("Invalid"), READY("Ready"), DISCONNECTED("Disconnected");

        private final String csName;

        private ClientState(String csName) {
            this.csName = csName;
        }

        public String getCsName() {
            return csName;
        }

        public static ClientState parseCsName(String name) {
            for (ClientState value : values()) {
                if (Objects.equals(name, value.getCsName())) {
                    return value;
                }
            }
            return null;
        }
    }
}
