/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.json.JsonObject;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class GlobalSettings {

    private final UUID id = UUID.randomUUID();
    private final String name;
    private final String serverGroup;
    private final String mod;
    private final String modVersion;
    private final String modTitle;
    private final String modWebsite;
    private final String modIcon32;
    private MapInfo map;
    private int timestep = 40;
    private int orderLatency = 3;
    private int randomSeed = new Random().nextInt();
    private boolean allowSpectators = true;
    private boolean allowVersionMismatch = false;
    private boolean enableSingleplayer = true;
    private final Map<String, Option> options = new HashMap<>();
    private String password;

    public GlobalSettings(String name, String serverGroup, JsonObject config) {
        this.name = name;
        this.serverGroup = serverGroup;
        this.mod = config.getString("mod");
        this.modVersion = config.getString("modversion");
        this.modTitle = config.getString("modtitle");
        this.modWebsite = config.getString("modwebsite");
        this.modIcon32 = config.getString("modicon32");
        this.map = new MapInfo(config.getString("map"), new JsonObject().put("Players", new JsonObject()));
        this.timestep = config.getInteger("timestep");
        this.orderLatency = config.getInteger("orderlatency");
        this.allowSpectators = config.getBoolean("allowspectators");
        this.allowVersionMismatch = config.getBoolean("allowversionmismatch");
        this.enableSingleplayer = config.getBoolean("enablesingleplayer");
        this.password = config.getString("password");

        for (Map.Entry<String, Object> option : config.getJsonObject("options")) {
            options.put(option.getKey(), new Option((JsonObject) option.getValue()));
        }
    }

    public JsonObject serialize() {
        JsonObject obj = new JsonObject();
        obj.put("GameUid", id.toString());
        obj.put("ServerName", name);
        obj.put("Map", map == null ? "" : map.getId());
        obj.put("Timestep", timestep);
        obj.put("OrderLatency", orderLatency);
        obj.put("RandomSeed", randomSeed);
        obj.put("AllowSpectators", allowSpectators ? "True" : "False");
        obj.put("AllowVersionMismatch", allowVersionMismatch ? "True" : "False");
        obj.put("EnableSingleplayer", enableSingleplayer ? "True" : "False");

        JsonObject optionsObj = new JsonObject();
        for (Map.Entry<String, Option> e : options.entrySet()) {
            optionsObj.put(e.getKey(), new JsonObject().put("IsLocked", e.getValue().isLocked() ? "True" : "False").put("Value", e.getValue().getValue()).put("PreferredValue", e.getValue().getPreferredValue()));
        }
        obj.put("Options", optionsObj);

        return obj;
    }

    public UUID getId() {
        return id;
    }

    public String getName() {
        return name;
    }

    public String getServerGroup() {
        return serverGroup;
    }

    public String getMod() {
        return mod;
    }

    public String getModVersion() {
        return modVersion;
    }

    public String getModTitle() {
        return modTitle;
    }

    public String getModWebsite() {
        return modWebsite;
    }

    public String getModIcon32() {
        return modIcon32;
    }

    public MapInfo getMap() {
        return map;
    }

    public void setMap(MapInfo map) {
        this.map = map;
    }

    public int getTimestep() {
        return timestep;
    }

    public void setTimestep(int timestep) {
        this.timestep = timestep;
    }

    public int getOrderLatency() {
        return orderLatency;
    }

    public void setOrderLatency(int orderLatency) {
        this.orderLatency = orderLatency;
    }

    public int getRandomSeed() {
        return randomSeed;
    }

    public void setRandomSeed(int randomSeed) {
        this.randomSeed = randomSeed;
    }

    public boolean isAllowSpectators() {
        return allowSpectators;
    }

    public void setAllowSpectators(boolean allowSpectators) {
        this.allowSpectators = allowSpectators;
    }

    public boolean isAllowVersionMismatch() {
        return allowVersionMismatch;
    }

    public void setAllowVersionMismatch(boolean allowVersionMismatch) {
        this.allowVersionMismatch = allowVersionMismatch;
    }

    public boolean isEnableSingleplayer() {
        return enableSingleplayer;
    }

    public void setEnableSingleplayer(boolean enableSingleplayer) {
        this.enableSingleplayer = enableSingleplayer;
    }

    public Map<String, Option> getOptions() {
        return options;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public class Option {

        private boolean locked;
        private String value;
        private String preferredValue;

        public Option(JsonObject obj) {
            locked = obj.getBoolean("locked");
            if (obj.getValue("value") instanceof Boolean) {
                value = obj.getBoolean("value") ? "True" : "False";
            } else {
                value = obj.getValue("value").toString();
            }
            if (obj.getValue("preferredvalue") instanceof Boolean) {
                preferredValue = obj.getBoolean("preferredvalue") ? "True" : "False";
            } else {
                preferredValue = obj.getValue("preferredvalue").toString();
            }
        }

        public boolean isLocked() {
            return locked;
        }

        public void setLocked(boolean locked) {
            this.locked = locked;
        }

        public String getValue() {
            return value;
        }

        public void setValue(String value) {
            this.value = value;
        }

        public String getPreferredValue() {
            return preferredValue;
        }

        public void setPreferredValue(String preferredValue) {
            this.preferredValue = preferredValue;
        }
    }
}
