/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.Handler;
import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpClientOptions;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import java.text.MessageFormat;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.as205315.applications.ora.server.utils.MiniYamlParser;
import net.as205315.applications.ora.server.utils.VertxUtils;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class MapManager {

    private static final Logger LOG = Logger.getLogger(MapManager.class.getName());
    private static final String RESOURCE_CENTER = "http://resource.openra.net/map/url/%s";

    private final Vertx vertx;
    private final HttpClient client;

    public MapManager(Vertx vertx) {
        this.vertx = vertx;
        client = vertx.createHttpClient(new HttpClientOptions());
    }

    public void getInfo(String id, Handler<MapInfo> callback, Handler<Throwable> error) {
        VertxUtils.getNow(client, String.format(RESOURCE_CENTER, id), res -> {
            if (res.statusCode() == 200) {
                Buffer buffer = Buffer.buffer();
                res.handler(buffer::appendBuffer);
                res.endHandler(v -> {
                    try {
                        JsonArray maps = new JsonArray(buffer);
                        for (Object map : maps) {
                            if (map instanceof JsonObject) {
                                String hash = ((JsonObject) map).getString("map_hash");
                                VertxUtils.getNow(client, ((JsonObject) map).getString("url").replace("/oramap", "/yaml"), yres -> {
                                    if (res.statusCode() == 200) {
                                        Buffer ybuffer = Buffer.buffer();
                                        yres.handler(ybuffer::appendBuffer);
                                        yres.endHandler(yv -> {
                                            try {
                                                callback.handle(parseYaml(hash, ybuffer.toString()));
                                            } catch (Throwable ex) {
                                                LOG.log(Level.SEVERE, "Exception while querying Resource Renter.", ex);
                                                error.handle(ex);
                                            }
                                        });
                                    }
                                });
                            } else {
                                String message = MessageFormat.format("Failed to request map yaml from Resource Center: HTTP-{0} ({1}).", res.statusCode(), res.statusMessage());
                                LOG.log(Level.WARNING, message);
                                error.handle(new RuntimeException(message));
                            }
                        }
                    } catch (Throwable ex) {
                        LOG.log(Level.SEVERE, "Exception while querying Resource Center.", ex);
                        error.handle(ex);
                    }
                });
            } else {
                String message = MessageFormat.format("Failed to request map info from Resource Center: HTTP-{0} ({1}).", res.statusCode(), res.statusMessage());
                LOG.log(Level.WARNING, message);
                error.handle(new RuntimeException(message));
            }
        });
    }

    private MapInfo parseYaml(String hash, String yaml) {
        try {
            return new MapInfo(hash, MiniYamlParser.parse(yaml));
        } catch (Throwable ex) {
            LOG.log(Level.SEVERE, "Exception while parsing map yaml.", ex);
            throw ex;
        }
    }
}
