/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpClientOptions;
import io.vertx.core.json.JsonObject;
import java.util.List;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.utils.MiniYamlParser;
import net.as205315.applications.ora.server.utils.VertxUtils;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class MasterServerPinger {

    private static final Logger LOG = Logger.getLogger(MasterServerPinger.class.getName());
    private static final String URI_PING = "http://master.openra.net/ping";

    private final Vertx vertx;
    private final HttpClient client;
    private final int port;

    public MasterServerPinger(Vertx vertx, int port) {
        this.vertx = vertx;
        this.client = vertx.createHttpClient(new HttpClientOptions());
        this.port = port;
    }

    public void updateState(GlobalSettings settings, Supplier<Stream<ClientInfo>> clientsAndBots, boolean started) {
        long clients = clientsAndBots.get()
                .filter(c -> (c.getBot() == null))
                .count();
        if (clients != 0) {
            LOG.log(Level.INFO, "{0} pings master server. {1} clients are currently connected in state {2}.", new Object[]{settings.getName(), clients, started ? "PLAYING" : "LOBBY"});
        }

        PingState state = new PingState();
        state.setProtocol(2);
        state.setName(settings.getName());
        state.setAddress("0.0.0.0:" + port);
        state.setMod(settings.getMod());
        state.setVersion(settings.getModVersion());
        state.setModTitle(settings.getModTitle());
        state.setModWebsite(settings.getModWebsite());
        state.setModIcon32(settings.getModIcon32());
        state.setMap(settings.getMap().getId());
        state.setState(started ? 2 : 1);
        state.setMaxPlayers(settings.getMap().getSlots().stream()
                .filter(MapInfo.Slot::isPlayable)
                .count());
        state.setIsProtected(settings.getPassword() != null);
        state.setClients(clientsAndBots.get()
                .filter(c -> c.getBot() == null));

        VertxUtils.putNow(client, URI_PING, MiniYamlParser.print(state.serialize()), res -> {
            if (res.statusCode() == 200) {
                Buffer buffer = Buffer.buffer();
                res.handler(buffer::appendBuffer);
                res.endHandler(v -> {
                    try {
                        Matcher matcher = Pattern.compile("^\\[(\\d+)\\]\\s*(.*)$").matcher(buffer.toString());
                        if (matcher.matches()) {
                            LOG.log(Level.WARNING, "Failed to ping master server: ORA-{0} ({1}).", new Object[]{matcher.group(1), matcher.group(2)});
                        }

                    } catch (Throwable ex) {
                        LOG.log(Level.SEVERE, "Exception while pinging master server.", ex);
                    }
                });
            } else {
                LOG.log(Level.WARNING, "Failed to ping master server: HTTP-{0} ({1}).", new Object[]{res.statusCode(), res.statusMessage()});
            }
        });
    }

    private class PingState {

        private int protocol;
        private String name;
        private String address;
        private String mod;
        private String version;
        private String modTitle;
        private String modWebsite;
        private String modIcon32;
        private String map;
        private int state;
        private long maxPlayers;
        private boolean isProtected;
        private List<ClientInfo> clients;

        public int getProtocol() {
            return protocol;
        }

        public void setProtocol(int protocol) {
            this.protocol = protocol;
        }

        public String getName() {
            return name;
        }

        public void setName(String name) {
            this.name = name;
        }

        public String getAddress() {
            return address;
        }

        public void setAddress(String address) {
            this.address = address;
        }

        public String getMod() {
            return mod;
        }

        public void setMod(String mod) {
            this.mod = mod;
        }

        public String getVersion() {
            return version;
        }

        public void setVersion(String version) {
            this.version = version;
        }

        public String getModTitle() {
            return modTitle;
        }

        public void setModTitle(String modTitle) {
            this.modTitle = modTitle;
        }

        public String getModWebsite() {
            return modWebsite;
        }

        public void setModWebsite(String modWebsite) {
            this.modWebsite = modWebsite;
        }

        public String getModIcon32() {
            return modIcon32;
        }

        public void setModIcon32(String modIcon32) {
            this.modIcon32 = modIcon32;
        }

        public String getMap() {
            return map;
        }

        public void setMap(String map) {
            this.map = map;
        }

        public int getState() {
            return state;
        }

        public void setState(int state) {
            this.state = state;
        }

        public long getMaxPlayers() {
            return maxPlayers;
        }

        public void setMaxPlayers(long maxPlayers) {
            this.maxPlayers = maxPlayers;
        }

        public boolean isIsProtected() {
            return isProtected;
        }

        public void setIsProtected(boolean isProtected) {
            this.isProtected = isProtected;
        }

        public Stream<ClientInfo> getClients() {
            return clients.stream();
        }

        public void setClients(Stream<ClientInfo> clients) {
            this.clients = clients.collect(Collectors.toList());
        }

        public JsonObject serialize() {
            JsonObject obj = new JsonObject();

            obj.put("Protocol", protocol);
            obj.put("Name", name);
            obj.put("Address", address);
            obj.put("Mod", mod);
            obj.put("Version", version);
            obj.put("ModTitle", modTitle);
            obj.put("ModWebsite", modWebsite);
            obj.put("ModIcon32", modIcon32);
            obj.put("Map", map);
            obj.put("State", state);
            obj.put("MaxPlayers", maxPlayers);
            obj.put("Protected", isProtected);

            JsonObject clientsObject = new JsonObject();
            obj.put("Clients", clientsObject);
            clients.forEach((c) -> {
                clientsObject.put("Client@" + c.getId(), c.serialize());
            });

            return new JsonObject().put("Game", obj);
        }
    }
}
