/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.AbstractVerticle;
import io.vertx.core.Handler;
import io.vertx.core.json.JsonObject;
import io.vertx.core.net.NetServer;
import io.vertx.core.net.NetSocket;
import java.time.Instant;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.packets.DisconnectPacket;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.packets.ServerErrorOrder;
import net.as205315.applications.ora.server.proxy.LobbyServerPacket;
import net.as205315.applications.ora.server.utils.PacketUtils;
import net.as205315.applications.ora.server.utils.SocketAddressParser;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class OraLobbyVerticle extends AbstractVerticle {

    private static final Logger LOG = Logger.getLogger(OraLobbyVerticle.class.getName());
    public static final String EVENTBUS_ADDRESS_ADD = OraLobbyVerticle.class.getName() + ".add";
    public static final String EVENTBUS_ADDRESS_REMOVE = OraLobbyVerticle.class.getName() + ".remove";
    private static final int ID_RANGE_SIZE = (int) Math.pow(2, 24);

    private final Set<OraProxy> proxies = new HashSet<>();
    private final Map<Integer, ServerState> serverStates = new HashMap<>();
    private NetServer server;

    @Override
    public void start() throws Exception {
        server = vertx.createNetServer();
        server.connectHandler(connectHandler());
        server.listen(config().getInteger(Config.CONFIG_LOBBY_PORT, 0), config().getString(Config.CONFIG_LOBBY_ADDRESS, "::"), res -> {
            if (res.succeeded()) {
                LOG.log(Level.INFO, "Lobby server \"{0}\" is listening on port {1}.", new Object[]{config().getString(Config.CONFIG_LOBBY_NAME), res.result().actualPort()});

                vertx.eventBus().consumer(EVENTBUS_ADDRESS_ADD, message -> {
                    try {
                        JsonObject request = ((JsonObject) message.body());
                        int id = request.getInteger(Config.CONFIG_CLIENT_PORT);
                        proxies.forEach((p) -> {
                            serverStates.put(id, new ServerState(id, vertx, request));
                            p.write(new LobbyServerPacket.StartLobby(id, p.getIdBase(), p.getIdLimit()));
                        });
                        message.reply(true);
                    } catch (RuntimeException ex) {
                        LOG.log(Level.SEVERE, "Failed to stop lobby server " + message.body() + ".", ex);
                        message.fail(0, ex.getMessage());
                    }
                });
                vertx.eventBus().consumer(EVENTBUS_ADDRESS_REMOVE, message -> {
                    try {
                        JsonObject request = ((JsonObject) message.body());
                        int id = request.getInteger(Config.CONFIG_CLIENT_PORT);
                        proxies.forEach(p -> p.write(new LobbyServerPacket.StopLobby(id)));
                        ServerState state = serverStates.remove(id);
                        state.stop();
                        message.reply(true);
                    } catch (RuntimeException ex) {
                        LOG.log(Level.SEVERE, "Failed to stop lobby server " + message.body() + ".", ex);
                        message.fail(0, ex.getMessage());
                    }
                });
            } else {
                LOG.log(Level.SEVERE, "Lobby server \"" + config().getString(Config.CONFIG_LOBBY_NAME) + "\" failed to listen.", res.cause());
            }
        });
    }

    @Override
    public void stop() throws Exception {
        server.close(v -> {
            LOG.log(Level.INFO, "Lobby server \"{0}\" stopped.", new Object[]{config().getString(Config.CONFIG_LOBBY_NAME)});
        });
    }

    private Handler<NetSocket> connectHandler() {
        return socket -> {
            int id = IntStream.rangeClosed(1, Integer.MAX_VALUE / ID_RANGE_SIZE)
                    .filter(i -> proxies.stream().noneMatch(p -> p.getIdBase() / ID_RANGE_SIZE == i))
                    .findFirst().orElseThrow(() -> new IllegalStateException("Out of id ranges!"));
            OraProxy proxy = new OraProxy(socket, id * ID_RANGE_SIZE, (id + 1) * ID_RANGE_SIZE);
            proxies.add(proxy);
            socket.handler(PacketUtils.createReader(LobbyServerPacket::fromBuffer, dataHandler(proxy), false, 0));
            socket.closeHandler(disconnectHandler(proxy));
        };
    }

    private Handler<Void> disconnectHandler(OraProxy proxy) {
        return v -> {
            proxies.remove(proxy);
            proxy.getClients().collect(Collectors.toSet()).forEach(proxy::removeClient);
        };
    }

    private Consumer<LobbyServerPacket> dataHandler(OraProxy proxy) {
        return packet -> {
            Optional<Client> client = Optional.empty();
            try {
                switch (packet.getType()) {
                    case STATE:
                        LobbyServerPacket.State proxyState = (LobbyServerPacket.State) packet;
                        proxy.setState(proxyState);
                        proxies.stream()
                                .filter(p -> p.getState() != null && p.getState().getOwnEndpoint() != null)
                                .filter(p -> !Objects.equals(proxyState.getOwnEndpoint(), p.getState().getOwnEndpoint()))
                                .filter(p -> !proxyState.getConnectedProxies().contains(p.getState().getOwnEndpoint()))
                                .forEach(p -> proxy.write(new LobbyServerPacket.ConnectTo(p.getState().getOwnEndpoint())));

                        serverStates.keySet().stream()
                                .filter(s -> !proxyState.getRunningLobbies().contains(s))
                                .forEach(s -> proxy.write(new LobbyServerPacket.StartLobby(s, proxy.getIdBase(), proxy.getIdLimit())));
                        proxyState.getRunningLobbies().stream()
                                .filter(s -> !serverStates.containsKey(s))
                                .forEach(s -> proxy.write(new LobbyServerPacket.StopLobby(s)));
                        break;
                    case CLIENT_CONNECTED:
                        ServerState serverState = serverStates.get(((LobbyServerPacket.ClientConnectedPacket) packet).getLobby());
                        LobbyServerPacket.ClientConnectedPacket connectedPacket = ((LobbyServerPacket.ClientConnectedPacket) packet);
                        client = Optional.of(proxy.addClient(new Client(serverState, proxy, connectedPacket.getClient(), SocketAddressParser.fromString(connectedPacket.getRemoteAddress()))));
                        client.ifPresent(c -> serverState.connected(c));
                        break;
                    case CLIENT_PACKET:
                        serverState = serverStates.get(((LobbyServerPacket.ClientPacket) packet).getLobby());
                        Packet oraPacket = PacketDeserializer.deserialize(((LobbyServerPacket.ClientPacket) packet).getPacket());
                        int clientId = oraPacket.getClient();
                        client = serverState.getClient(clientId);

                        if (oraPacket instanceof DisconnectPacket) {
                            client.ifPresent(c -> serverState.disconnected(c));
                            client.ifPresent(c -> proxy.removeClient(c));
                        } else {
                            client.ifPresent(c -> {
                                if (oraPacket.getFrame() != 0) {
                                    c.setLastFrame(oraPacket.getFrame());
                                }
                                c.setLastPacket(Instant.now());
                            });
                            client.ifPresent(c -> serverState.received(c, oraPacket));
                            client.orElseThrow(() -> new IllegalStateException("Client packet received before client connected packet."));
                        }
                        break;
                    default:
                        LOG.log(Level.SEVERE, "Packet type {0} can''t be handled.", packet.getType());
                }
            } catch (Exception ex) {
                LOG.log(Level.SEVERE, "Exception while receiving and processing packet from proxy " + proxy.getSocket().remoteAddress() + client.map(c -> " (clientId=" + c.getId() + ")").orElse("") + ".", ex);
                client.ifPresent(c -> c.sendPacket(new ServerErrorOrder(0, 0, ex.getClass().getName() + ": " + ex.getMessage())));
                client.ifPresent(c -> c.disconnect());
            }
        };
    }

    public void broadcast(LobbyServerPacket packet) {
        proxies.forEach(p -> p.write(packet));
    }
}
