/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.buffer.Buffer;
import net.as205315.applications.ora.server.packets.ChatOrder;
import net.as205315.applications.ora.server.packets.ClientOrder;
import net.as205315.applications.ora.server.packets.CommandOrder;
import net.as205315.applications.ora.server.packets.DisconnectPacket;
import net.as205315.applications.ora.server.packets.HandshakeResponseOrder;
import net.as205315.applications.ora.server.packets.MessageOrder;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.packets.PongOrder;
import net.as205315.applications.ora.server.packets.SyncPacket;
import net.as205315.applications.ora.server.packets.TeamChatOrder;
import net.as205315.applications.ora.server.utils.CSCompat;
import net.as205315.applications.ora.server.utils.MiniYamlParser;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class PacketDeserializer {

    public static Packet deserialize(Buffer buffer) {
        int client = buffer.getIntLE(0);
        int frame = buffer.getIntLE(4);

        Packet packet;

        if (buffer.length() < 9) {
            packet = new ClientOrder(client, frame, Buffer.buffer());
        } else {
            switch (buffer.getUnsignedByte(8)) {
                // Client order.
                case 0xFF:
                    packet = new ClientOrder(client, frame, buffer.slice(8, buffer.length()));
                    break;
                // Server order.
                case 0xFE:
                    int nameStart = 9 + CSCompat.read7BitIntLength(buffer, 9);
                    int nameLength = CSCompat.read7BitInt(buffer, 9);
                    String name = buffer.getString(nameStart, nameStart + nameLength);
                    int dataStart = nameStart + nameLength + CSCompat.read7BitIntLength(buffer, nameStart + nameLength);
                    int dataLength = CSCompat.read7BitInt(buffer, nameStart + nameLength);
                    String data = buffer.getString(dataStart, dataStart + dataLength);

                    if (dataStart + dataLength < buffer.length()) {
                        throw new IllegalArgumentException("There are " + (buffer.length() - dataStart - dataLength) + " bytes left in the packet after the end of the ServerOrder.");
                    }

                    switch (name) {
                        case "HandshakeResponse":
                            packet = new HandshakeResponseOrder(client, frame, MiniYamlParser.parse(data));
                            break;
                        case "Chat":
                            packet = new ChatOrder(client, frame, data);
                            break;
                        case "TeamChat":
                            packet = new TeamChatOrder(client, frame, data);
                            break;
                        case "Command":
                            packet = new CommandOrder(client, frame, data);
                            break;
                        case "Message":
                            packet = new MessageOrder(client, frame, data);
                            break;
                        case "Pong":
                            packet = new PongOrder(client, frame, Long.valueOf(data));
                            break;
                        default:
                            throw new UnsupportedOperationException("Unknown server order: " + name);
                    }
                    break;
                // Disconnect packet.
                case 0xBF:
                    packet = new DisconnectPacket(client, frame);
                    break;
                // Sync packet.
                case 0x65:
                    packet = new SyncPacket(client, frame, buffer.slice(9, buffer.length()));
                    break;
                default:
                    throw new UnsupportedOperationException("Unknown packet: " + buffer.getUnsignedByte(8));
            }
        }

        return packet;
    }
}
