/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.lobby;

import io.vertx.core.Vertx;
import io.vertx.core.json.JsonObject;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.logic.LobbyServerLogic;
import net.as205315.applications.ora.server.logic.ServerLogic;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.packets.ServerErrorOrder;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ServerState {

    private static final Logger LOG = Logger.getLogger(ServerState.class.getName());

    private final Vertx vertx;
    private final JsonObject config;
    private final long timer;
    private final List<Client> clients = new ArrayList<>();
    private final int id;
    private ServerLogic logic;

    public ServerState(int port, Vertx vertx, JsonObject config) {
        this.vertx = vertx;
        this.config = config;
        this.id = port;
        logic = new LobbyServerLogic(this, config, new MasterServerPinger(vertx, port), Collections.unmodifiableList(clients));
        timer = vertx.setPeriodic(10 * 1000, this::tick);

        LOG.log(Level.INFO, "Lobby {0} ({1}) started successfully.", new Object[]{id, config.getString(Config.CONFIG_LOBBY_NAME)});
    }

    public void stop() {
        for (Client client : clients) {
            try {
                client.sendPacket(new ServerErrorOrder(0, 0, "Server is shutting down."));
            } catch (Throwable ex) {
                LOG.log(Level.WARNING, "Exception while sending disconnect packet to client " + client.getClientInfo().getAddress() + ".", ex);
            }
        }

        vertx.cancelTimer(timer);

        LOG.log(Level.INFO, "Lobby {0} ({1}) stopped successfully.", new Object[]{id, config.getString(Config.CONFIG_LOBBY_NAME)});
    }

    public int getId() {
        return id;
    }

    private void changeLogicIfNeeded(ServerLogic newLogic) {
        if (logic != Objects.requireNonNull(newLogic, "newLogic")) {
            logic = newLogic;
        }
    }

    public void connected(Client client) {
        try {
            clients.add(client);
            changeLogicIfNeeded(logic.connected(client));
        } catch (RuntimeException ex) {
            LOG.log(Level.WARNING, "Exception while processing client connected event.", ex);
            client.sendPacket(new ServerErrorOrder(client.getId(), 0, ex.getMessage()));
            client.disconnect();
        }
    }

    public void received(Client client, Packet packet) {
        try {
            changeLogicIfNeeded(logic.received(client, packet));
        } catch (RuntimeException ex) {
            LOG.log(Level.WARNING, "Exception while processing client packet.", ex);
            client.sendPacket(new ServerErrorOrder(client.getId(), 0, ex.getMessage()));
            client.disconnect();
        }
    }

    public void disconnected(Client client) {
        try {
            clients.remove(client);
            changeLogicIfNeeded(logic.disconnected(client));
        } catch (RuntimeException ex) {
            LOG.log(Level.WARNING, "Exception while processing client disconnected event.", ex);
            client.sendPacket(new ServerErrorOrder(client.getId(), 0, ex.getMessage()));
            client.disconnect();
        }
    }

    public void tick(long count) {
        changeLogicIfNeeded(logic.tick());
    }

    public Vertx getVertx() {
        return vertx;
    }

    public Optional<Client> getClient(int clientId) {
        return clients.stream().filter(c -> c.getId() == clientId).findAny();
    }
}
