/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.logic;

import io.vertx.core.buffer.Buffer;
import io.vertx.core.json.JsonObject;
import java.math.BigInteger;
import java.time.Duration;
import java.time.Instant;
import java.util.List;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.authentication.AuthenticationManager;
import net.as205315.applications.ora.server.commands.CommandManager;
import net.as205315.applications.ora.server.lobby.Client;
import net.as205315.applications.ora.server.lobby.ClientInfo;
import net.as205315.applications.ora.server.lobby.GlobalSettings;
import net.as205315.applications.ora.server.lobby.MasterServerPinger;
import net.as205315.applications.ora.server.lobby.ReplayManager;
import net.as205315.applications.ora.server.lobby.ServerState;
import net.as205315.applications.ora.server.packets.ChatOrder;
import net.as205315.applications.ora.server.packets.ClientOrder;
import net.as205315.applications.ora.server.packets.DisconnectPacket;
import net.as205315.applications.ora.server.packets.MessageOrder;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.packets.PingOrder;
import net.as205315.applications.ora.server.packets.PongOrder;
import net.as205315.applications.ora.server.packets.ReplayMetadataOrder;
import net.as205315.applications.ora.server.packets.ServerErrorOrder;
import net.as205315.applications.ora.server.packets.ServerOrder;
import net.as205315.applications.ora.server.packets.StartGameOrder;
import net.as205315.applications.ora.server.packets.SyncClientPingsOrder;
import net.as205315.applications.ora.server.packets.SyncInfoOrder;
import net.as205315.applications.ora.server.packets.SyncPacket;
import net.as205315.applications.ora.server.packets.TeamChatOrder;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class IngameServerLogic extends ServerLogic {

    private static final Duration GRACE_PERIOD = Duration.ofSeconds(10);

    private final ServerState serverState;
    private final MasterServerPinger pinger;
    private final ReplayManager replayManager;
    private final GlobalSettings settings;
    private final AuthenticationManager authenticationManager;
    private final CommandManager commandManager;
    private final Buffer replay;
    private final Instant start;
    private int pingerCounter = 0;

    public IngameServerLogic(ServerState serverState, JsonObject config, MasterServerPinger pinger, List<Client> clients, GlobalSettings settings) {
        super(serverState, config, clients);
        this.serverState = serverState;
        this.pinger = pinger;
        this.replayManager = new ReplayManager(vertx());
        this.settings = settings;
        this.authenticationManager = new AuthenticationManager(settings, vertx(), this::getAllClients);
        this.commandManager = new CommandManager(clients()::stream, authenticationManager, settings);

        this.replay = Buffer.buffer();

        writeReplay(new SyncInfoOrder(0, 0, settings, clients.stream().map(Client::getClientInfo)));
        clients().forEach(client -> {
            client.setLastFrame(0);
            client.setLastPacket(Instant.now());

            // Disconnect all spectators so they can't freeze the game.
            if (isSpectator(client)) {
                client.setAllowDirect(false);
                broadcast(new DisconnectPacket(client.getId(), 0x0));
            } else {
                client.setAllowDirect(true);
                broadcast(new DisconnectPacket(client.getId(), 0x7FFFFFFF));
            }

            writeReplay(new DisconnectPacket(client.getId(), 0x7FFFFFFF));
        });
        broadcastWithReplay(new StartGameOrder(0, 0));
        start = Instant.now();
    }

    @Override
    public ServerLogic connected(Client client) {
        client.sendPacket(new ServerErrorOrder(0, 0, "Game is already running."));
        client.disconnect();
        return this;
    }

    @Override
    public ServerLogic received(Client client, Packet packet) {
        if (packet instanceof ServerOrder && packet.getFrame() != 0) {
            client.sendPacket(new ServerErrorOrder(0, 0, "Received " + ((ServerOrder) packet).getOrderName() + " attached to a tick. Please apply the patch from https://github.com/OpenRA/OpenRA/pull/16052. There is no safe way to continue the game without kicking you."));
            client.disconnect();
            if (!isSpectator(client)) {
                broadcastNoLoop(new MessageOrder(client.getClientInfo().getName() + " sent a " + ((ServerOrder) packet).getOrderName() + " attached to a tick. There is no safe way to continue the game without kicking him."));
                broadcastNoLoop(new ClientOrder(client.getId(), packet.getFrame(), Buffer.buffer()));
            }
            return this;
        }

        if (packet instanceof ClientOrder || packet instanceof SyncPacket) {
            // Send packet only to spectators except those from spectators. There is no need to send packets from players to players because that's already handled by the proxies.
            if (!isSpectator(client)) {
                clients().stream()
                        .filter(isSpectator())
                        .filter(c -> c != client)
                        .forEach(c -> {
                            c.sendPacket(packet);
                        });
            }
            writeReplay(packet);
        } else if (packet instanceof PongOrder) {
            BigInteger now = new BigInteger(Long.toUnsignedString(System.nanoTime()));
            BigInteger last = new BigInteger(Long.toUnsignedString(((PongOrder) packet).getTimestamp()));
            client.getClientInfo().appendLatency(now.subtract(last).longValue());
        } else if (packet instanceof ChatOrder) {
            if (!commandManager.handleMessage(client, (ChatOrder) packet)) {
                broadcastNoLoop(packet);
                writeReplay(packet);
            }
        } else if (packet instanceof TeamChatOrder) {
            clients().stream()
                    .filter(c -> !client.equals(c))
                    .filter(isSpectatorOrSameTeamAs(client))
                    .forEach(c -> c.sendPacket(packet));
            writeReplay(packet);
        } else {
            throw new UnsupportedOperationException("Unknown packet " + packet.getClass().getName() + ".");
        }

        return this;
    }

    @Override
    public ServerLogic disconnected(Client client) {
        if (client.getClientInfo() != null) {
            broadcastWithReplay(new MessageOrder(client.getClientInfo().getName() + " disconnected."));
            if (!isSpectator(client)) {
                broadcastWithReplay(new DisconnectPacket(client.getId(), client.getLastFrame()));
            } else {
                writeReplay(new DisconnectPacket(client.getId(), client.getLastFrame()));
            }
        }

        // Go back to lobby if empty.
        if (clients().isEmpty()) {
            writeReplay(new ReplayMetadataOrder(0, 0, settings, start, Instant.now()));
            replayManager.saveReplay(settings, replay);

            return new LobbyServerLogic(serverState, config(), pinger, clients());
        } else {
            return this;
        }
    }

    @Override
    public ServerLogic tick() {
        broadcast(new SyncClientPingsOrder(0, 0, getAllClients()));
        broadcast(new PingOrder(0, 0, System.nanoTime()));
        clients().stream()
                .filter(isSpectator().negate())
                .filter(c -> c.getLastPacket().plus(GRACE_PERIOD).isBefore(Instant.now()))
                .forEach(c -> {
                    broadcastWithReplay(new MessageOrder(c.getClientInfo().getName() + " is not responding. Please use the command \"!kick " + c.getClientInfo().getName() + "\" if you want to kick him manually."));
                });

        if (pingerCounter % 12 == 0) {
            pinger.updateState(settings, ()
                    -> clients().stream()
                            .filter(c -> c.getClientInfo() != null)
                            .map(Client::getClientInfo),
                    true);
        }
        pingerCounter++;

        return this;
    }

    private void writeReplay(Packet packet) {
        replay.appendBuffer(packet.toReplayBuffer());
    }

    private void broadcastWithReplay(Packet packet) {
        broadcast(packet);
        writeReplay(packet);
    }

    private Stream<ClientInfo> getAllClients() {
        return clients().stream()
                .filter(c -> c.getClientInfo() != null)
                .map(Client::getClientInfo);
    }
}
