/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.logic;

import io.vertx.core.json.JsonObject;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Random;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.authentication.AuthenticationManager;
import net.as205315.applications.ora.server.authentication.ClientAuthentication;
import net.as205315.applications.ora.server.commands.CommandManager;
import net.as205315.applications.ora.server.lobby.Client;
import net.as205315.applications.ora.server.lobby.ClientInfo;
import net.as205315.applications.ora.server.lobby.GlobalSettings;
import net.as205315.applications.ora.server.lobby.MapInfo;
import net.as205315.applications.ora.server.lobby.MapManager;
import net.as205315.applications.ora.server.lobby.MasterServerPinger;
import net.as205315.applications.ora.server.lobby.ServerState;
import net.as205315.applications.ora.server.packets.ChatOrder;
import net.as205315.applications.ora.server.packets.CommandOrder;
import net.as205315.applications.ora.server.packets.DisconnectPacket;
import net.as205315.applications.ora.server.packets.HandshakeRequestOrder;
import net.as205315.applications.ora.server.packets.HandshakeResponseOrder;
import net.as205315.applications.ora.server.packets.MessageOrder;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.packets.PingOrder;
import net.as205315.applications.ora.server.packets.PongOrder;
import net.as205315.applications.ora.server.packets.ServerErrorOrder;
import net.as205315.applications.ora.server.packets.SyncClientPingsOrder;
import net.as205315.applications.ora.server.packets.SyncInfoOrder;
import net.as205315.applications.ora.server.packets.TeamChatOrder;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class LobbyServerLogic extends ServerLogic {

    private static final int BOT_ID_START = 100000000;

    private final MapManager mapCache;
    private final GlobalSettings settings;
    private final MasterServerPinger pinger;
    private final AuthenticationManager authenticationManager;
    private final CommandManager commandManager;
    private final List<ClientInfo> bots = new ArrayList<>();
    private int pingerCounter = 0;

    public LobbyServerLogic(ServerState serverState, JsonObject config, MasterServerPinger pinger, List<Client> clients) {
        super(serverState, config, clients);
        clients.forEach(Client::disconnect);

        this.mapCache = new MapManager(vertx());
        this.pinger = pinger;
        this.settings = new GlobalSettings(config.getString(Config.CONFIG_LOBBY_NAME), config.getString(Config.CONFIG_LOBBY_GROUP), config.getJsonObject(Config.CONFIG_LOBBY_CONFIG));
        this.authenticationManager = new AuthenticationManager(settings, vertx(), this::getAllClients);
        this.commandManager = new CommandManager(clients()::stream, authenticationManager, settings);

        setMap(null, settings.getMap().getId(), true);
    }

    @Override
    public ServerLogic connected(Client client) {
        client.sendPacket(new HandshakeRequestOrder(0, 0, settings.getMod(), settings.getModVersion(), settings.getMap().getId()));
        return this;
    }

    @Override
    public ServerLogic received(Client client, Packet packet) {
        if (packet instanceof HandshakeResponseOrder) {
            client.setClientInfo(((HandshakeResponseOrder) packet).getClientInfo());
            client.getClientInfo().setClientAuthentication(authenticationManager.authenticate(client.getClientInfo().getName(), ((HandshakeResponseOrder) packet).getPassword(), client.getRemoteAddress().getAddress().getHostAddress(), client.getClientInfo()));

            if (!client.getClientInfo().getClientAuthentication().requirePermissionOrDisconnect(client, ClientAuthentication.Permission.CONNECT)) {
                return this;
            }

            client.getClientInfo().setId(client.getId());
            client.getClientInfo().setAddress(client.getRemoteAddress().getAddress().getHostAddress());
            broadcastLobbyState();
            broadcast(new MessageOrder(client.getClientInfo().getName() + " joined. "));
            return this;
        }

        if (client.getClientInfo() == null) {
            throw new IllegalStateException("Handshake is not completed.");
        }

        if (packet instanceof PongOrder) {
            BigInteger now = new BigInteger(Long.toUnsignedString(System.nanoTime()));
            BigInteger last = new BigInteger(Long.toUnsignedString(((PongOrder) packet).getTimestamp()));
            client.getClientInfo().appendLatency(now.subtract(last).longValue());
        } else if (packet instanceof ChatOrder) {
            if (commandManager.handleMessage(client, (ChatOrder) packet)) {
                broadcastLobbyState();
            } else {
                broadcastNoLoop(packet);
            }
        } else if (packet instanceof TeamChatOrder) {
            clients().stream()
                    .filter(c -> !client.equals(c))
                    .filter(isSpectatorOrSameTeamAs(client))
                    .forEach(c -> c.sendPacket(packet));
        } else if (packet instanceof CommandOrder) {
            String[] parts = ((CommandOrder) packet).getMessage().split(" ", 2);
            switch (parts[0]) {
                case "state":
                    client.getClientInfo().setState(ClientInfo.ClientState.parseCsName(parts[1]));
                    break;
                case "map":
                    if (client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        setMap(client, parts[1], false);
                    }
                    break;
                case "slot":
                    if (getAllClientsAndBots()
                            .map(ClientInfo::getSlot)
                            .anyMatch(s -> Objects.equals(s, parts[1]))) {
                        return this;
                    }

                    MapInfo.Slot oldSlot = getSlot(client.getClientInfo().getSlot());
                    MapInfo.Slot newSlot = getSlot(parts[1]);

                    if (oldSlot != null && oldSlot.isLockSpawn()) {
                        client.getClientInfo().setSpawnPoint(0);
                    }

                    if (newSlot != null) {
                        if (newSlot.isLockSpawn()) {
                            client.getClientInfo().setSpawnPoint(newSlot.getSpawn());
                        }
                        if (newSlot.isLockFaction()) {
                            client.getClientInfo().setFaction(newSlot.getFaction());
                        } else if (client.getClientInfo().getFaction() == null) {
                            client.getClientInfo().setFaction("Random");
                        }
                        if (newSlot.isLockTeam()) {
                            client.getClientInfo().setTeam(newSlot.getTeam());
                        }
                        client.getClientInfo().setSlot(parts[1]);
                    }
                    break;
                case "spectate":
                    client.getClientInfo().setSlot(null);
                    client.getClientInfo().setSpawnPoint(0);
                    break;
                case "allow_spectators":
                    settings.setAllowSpectators(Boolean.parseBoolean(parts[1]));
                    break;
                case "slot_open":
                    MapInfo.Slot slot = Objects.requireNonNull(getSlot(parts[1]), "slot");
                    slot.setClosed(false);
                    bots.removeIf(b -> Objects.equals(b.getSlot(), slot.getId()));
                    getAllClients()
                            .filter(c -> Objects.equals(c.getSlot(), slot.getId()))
                            .forEach(c -> {
                                c.setSlot(null);
                                c.setSpawnPoint(0);
                            });
                    break;
                case "slot_close":
                    slot = Objects.requireNonNull(getSlot(parts[1]), "slot");
                    slot.setClosed(true);
                    bots.removeIf(b -> Objects.equals(b.getSlot(), slot.getId()));
                    getAllClients()
                            .filter(c -> Objects.equals(c.getSlot(), slot.getId()))
                            .forEach(c -> {
                                c.setSlot(null);
                                c.setSpawnPoint(0);
                            });
                    break;
                case "slot_bot":
                    String slotName = parts[1].split(" ", 3)[0];
                    String botName = parts[1].split(" ", 3)[2];
                    slot = Objects.requireNonNull(getSlot(slotName), "slot");
                    if (!slot.isAllowBots() || getAllClients().map(ClientInfo::getSlot).anyMatch(s -> Objects.equals(s, slotName))) {
                        return this;
                    }
                    Optional<ClientInfo> bot = getAllBots()
                            .filter(b -> Objects.equals(b.getSlot(), slotName))
                            .findAny();
                    if (bot.isPresent()) {
                        bot.get().setName(botName);
                        bot.get().setBot(botName);
                    } else {
                        ClientInfo b = new ClientInfo(getBotClientId());
                        b.setName(botName);
                        b.setBot(botName);
                        b.setSlot(slotName);
                        if (slot.isLockSpawn()) {
                            b.setSpawnPoint(slot.getSpawn());
                        }
                        if (slot.isLockFaction()) {
                            b.setFaction(slot.getFaction());
                        } else {
                            b.setFaction("Random");
                        }
                        if (slot.isLockTeam()) {
                            b.setTeam(slot.getTeam());
                        }
                        b.setState(ClientInfo.ClientState.READY);
                        b.setColor(String.format("%08X", 0xFF000000 | new Random().nextInt()).substring(2, 8));
                        b.setPreferredColor(b.getColor());
                        bots.add(Objects.requireNonNull(b, "b"));
                    }
                    break;
                case "kick":
                    int id = Integer.parseInt(parts[1].split(" ", 2)[0]);
                    clients().stream()
                            .filter(c -> c.getId() == id)
                            .forEach(c -> {
                                c.sendPacket(new ServerErrorOrder(client.getId(), 0, "You have been kicked."));
                                c.disconnect();
                            });
                    bots.removeIf(b -> b.getId() == id);
                    break;
                case "name":
                    client.getClientInfo().setName(parts[1]);
                    break;
                case "faction":
                    int targetId = Integer.parseInt(parts[1].split(" ", 2)[0]);
                    if (client.getId() != targetId && !client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        return this;
                    }

                    getAllClientsAndBots()
                            .filter(c -> c.getId() == targetId)
                            .findAny()
                            .ifPresent(c -> {
                                MapInfo.Slot s = getSlot(c.getSlot());
                                if (s != null && s.isLockFaction()) {
                                    c.setFaction(s.getFaction());
                                } else {
                                    c.setFaction(parts[1].split(" ", 2)[1]);
                                }
                            });
                    break;
                case "team":
                    targetId = Integer.parseInt(parts[1].split(" ", 2)[0]);
                    if (client.getId() != targetId && !client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        return this;
                    }

                    getAllClientsAndBots()
                            .filter(c -> c.getId() == targetId)
                            .findAny()
                            .ifPresent(c -> {
                                MapInfo.Slot s = getSlot(c.getSlot());
                                if (s != null && s.isLockTeam()) {
                                    c.setTeam(s.getTeam());
                                } else {
                                    c.setTeam(Integer.parseInt(parts[1].split(" ", 2)[1]));
                                }
                            });
                    break;
                case "spawn":
                    targetId = Integer.parseInt(parts[1].split(" ", 2)[0]);
                    if (client.getId() != targetId && !client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        return this;
                    }

                    getAllClientsAndBots()
                            .filter(c -> c.getId() == targetId)
                            .findAny()
                            .ifPresent(c -> {
                                MapInfo.Slot s = getSlot(c.getSlot());
                                if (s == null) {
                                    c.setSpawnPoint(0);
                                } else if (s.isLockSpawn()) {
                                    c.setSpawnPoint(s.getSpawn());
                                } else {
                                    int spawn = Integer.parseInt(parts[1].split(" ", 2)[1]);
                                    if (settings.getMap().getSlots().stream()
                                            .filter(MapInfo.Slot::isLockSpawn)
                                            .anyMatch(s2 -> s2.getSpawn() == spawn)) {
                                        return;
                                    }
                                    if (spawn != 0 && getAllClientsAndBots().anyMatch(other -> other.getSpawnPoint() == spawn)) {
                                        return;
                                    }

                                    c.setSpawnPoint(spawn);
                                }
                            });
                    break;
                case "color":
                    targetId = Integer.parseInt(parts[1].split(" ", 2)[0]);
                    if (client.getId() != targetId && !client.getClientInfo().getClientAuthentication().requirePermissionOrInform(client, ClientAuthentication.Permission.ADMIN)) {
                        return this;
                    }

                    getAllClientsAndBots()
                            .filter(c -> c.getId() == targetId)
                            .findAny()
                            .ifPresent(c -> {
                                MapInfo.Slot s = getSlot(c.getSlot());
                                if (s != null && s.isLockColor()) {
                                    c.setColor(s.getColor());
                                } else {
                                    c.setColor(parts[1].split(" ", 2)[1]);
                                }
                            });
                    break;
                case "option":
                    settings.getOptions().get(parts[1].split(" ", 2)[0]).setValue(parts[1].split(" ", 2)[1]);
                    break;
                case "assignteams":
                    // TODO: Implement this menu option?
                    broadcast(new MessageOrder("\"assignteams\" is not supported yet."));
                    break;
                case "sync_lobby":
                    // TODO: Restart command?
                    broadcast(new MessageOrder("\"sync_lobby\" is not supported yet."));
                    break;
                case "startgame":
                    broadcast(new MessageOrder("\"Start Game\" is not supported, the game will start automatically once all players are ready."));
                    break;
                default:
                    throw new UnsupportedOperationException("Unknown server command " + parts[0]);
            }

            broadcastLobbyState();
        } else {
            throw new UnsupportedOperationException("Unknown packet " + packet.getClass().getName() + ".");
        }

        return checkAutoStart();
    }

    @Override
    public ServerLogic disconnected(Client client) {
        if (client.getClientInfo() != null) {
            broadcastLobbyState();
            broadcast(new MessageOrder(client.getClientInfo().getName() + " left. "));
            broadcast(new DisconnectPacket(client.getId(), 0));
        }
        return this;
    }

    @Override
    public ServerLogic tick() {
        broadcast(new SyncClientPingsOrder(0, 0, getAllClients()));
        broadcast(new PingOrder(0, 0, System.nanoTime()));
        if (pingerCounter % 12 == 0) {
            pinger.updateState(settings, this::getAllClientsAndBots, false);
        }
        pingerCounter++;
        return this;
    }

    private void setMap(Client client, String id, boolean refresh) {
        if (refresh || settings.getMap() == null || !Objects.equals(id, settings.getMap().getId())) {
            mapCache.getInfo(id, map -> {
                settings.setMap(map);

                for (Client c : clients()) {
                    if (c.getClientInfo() != null) {
                        c.getClientInfo().setSlot(null);
                        c.getClientInfo().setSpawnPoint(0);
                    }
                }

                broadcastLobbyState();

                if (client != null) {
                    broadcast(new MessageOrder(client.getClientInfo().getName() + " changed the map to " + settings.getMap().getTitle() + " (" + settings.getMap().getId() + ")"));
                }
            }, ex -> {
                broadcast(new MessageOrder(ex.getMessage()));
            });
        }
    }

    private MapInfo.Slot getSlot(String id) {
        return (id == null)
                ? null
                : settings.getMap().getSlots().stream()
                        .filter(s -> Objects.equals(s.getId(), id))
                        .findAny()
                        .orElseThrow(() -> new IllegalArgumentException("Slot " + id + " does not exist!"));
    }

    private Stream<ClientInfo> getAllClients() {
        return clients().stream()
                .filter(c -> c.getClientInfo() != null)
                .map(Client::getClientInfo);
    }

    private Stream<ClientInfo> getAllBots() {
        return bots.stream();
    }

    private Stream<ClientInfo> getAllClientsAndBots() {
        return Stream.concat(getAllClients(), getAllBots());
    }

    private int getBotClientId() {
        return IntStream.iterate(BOT_ID_START, i -> i + 1)
                .filter(i -> getAllClientsAndBots()
                .map(ClientInfo::getId).noneMatch(ci -> ci == i))
                .findFirst()
                .orElseThrow(() -> new IllegalStateException("No free client ID available!"));
    }

    private void broadcastLobbyState() {
        broadcast(new SyncInfoOrder(0, 0, settings, getAllClientsAndBots()));
        broadcast(new SyncClientPingsOrder(0, 0, getAllClients()));
        pinger.updateState(settings, this::getAllClientsAndBots, false);
    }

    private ServerLogic checkAutoStart() {
        if (getAllClients()
                .filter(c -> c.getSlot() != null)
                .count() == 0) {
            return this;
        }
        if (getAllClientsAndBots()
                .filter(c -> c.getSlot() != null)
                .anyMatch(c -> c.getState() != ClientInfo.ClientState.READY)) {
            return this;
        }
        if (settings.getMap().getSlots().stream()
                .filter(MapInfo.Slot::isRequired)
                .anyMatch(s -> getAllClientsAndBots()
                .noneMatch(c -> Objects.equals(c.getSlot(), s.getId())))) {
            return this;
        }

        if (!settings.isAllowSpectators()) {
            clients().stream()
                    .filter(c -> c.getClientInfo().getSlot() == null)
                    .forEach(c -> {
                        c.sendPacket(new ServerErrorOrder(0, 0, "Spectators are disabled."));
                        c.disconnect();
                    });
        }

        int i = 0;
        List<ClientInfo> clients = getAllClients()
                .filter(c -> c.getSlot() != null)
                .collect(Collectors.toList());
        getAllBots().forEach((bot) -> {
            bot.setBotController(clients.get(i % clients.size()).getId());
        });

        broadcastLobbyState();
        pinger.updateState(settings, this::getAllClientsAndBots, true);
        return new IngameServerLogic(serverState(), config(), pinger, clients(), settings);
    }
}
