/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.logic;

import io.vertx.core.Vertx;
import io.vertx.core.json.JsonObject;
import java.util.List;
import java.util.function.Predicate;
import net.as205315.applications.ora.server.lobby.Client;
import net.as205315.applications.ora.server.lobby.ServerState;
import net.as205315.applications.ora.server.packets.Packet;
import net.as205315.applications.ora.server.proxy.LobbyServerPacket;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public abstract class ServerLogic {

    private final ServerState serverState;
    private final JsonObject config;
    private final List<Client> clients;

    public ServerLogic(ServerState serverState, JsonObject config, List<Client> clients) {
        this.serverState = serverState;
        this.config = config;
        this.clients = clients;
    }

    protected ServerState serverState() {
        return serverState;
    }

    protected JsonObject config() {
        return config;
    }

    protected Vertx vertx() {
        return serverState().getVertx();
    }

    protected List<Client> clients() {
        return clients;
    }

    protected void broadcast(Packet packet) {
        clients.stream()
                .map(Client::getProxy)
                .distinct()
                .forEach(p -> p.write(new LobbyServerPacket.ClientPacket(serverState.getId(), packet.toBuffer())));
    }

    protected void broadcastNoLoop(Packet packet) {
        clients.stream()
                .filter(c -> c.getId() != packet.getClient())
                .forEach(c -> c.sendPacket(packet));
    }

    public abstract ServerLogic connected(Client client);

    public abstract ServerLogic received(Client client, Packet packet);

    public abstract ServerLogic disconnected(Client client);

    public abstract ServerLogic tick();

    protected boolean isSpectator(Client client) {
        return client.getClientInfo() == null || client.getClientInfo().getSlot() == null;
    }

    protected Predicate<Client> isSpectator() {
        return this::isSpectator;
    }

    protected boolean isSpectatorOrSameTeamAs(Client client, Client other) {
        if (isSpectator(client)) {
            return true;
        } else {
            if (isSpectator(other)) {
                return false;
            } else if (other.getClientInfo().getTeam() == 0) {
                return false;
            } else {
                return other.getClientInfo().getTeam() == client.getClientInfo().getTeam();
            }
        }
    }

    protected Predicate<Client> isSpectatorOrSameTeamAs(Client other) {
        return client -> isSpectatorOrSameTeamAs(client, other);
    }
}
