/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.management;

import io.vertx.core.AbstractVerticle;
import io.vertx.core.AsyncResult;
import io.vertx.core.Handler;
import io.vertx.core.eventbus.Message;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.handler.BodyHandler;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.lobby.OraLobbyVerticle;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ManagementVerticle extends AbstractVerticle {

    private static final Logger LOG = Logger.getLogger(OraLobbyVerticle.class.getName());

    private HttpServer server;
    private Router router;
    private final Map<String, JsonObject> state = new HashMap<>();

    @Override
    public void start() throws Exception {
        router = Router.router(vertx);

        router.put().path("/:group/:name").handler(BodyHandler.create()).handler(c -> {
            String name = c.pathParam("name");
            String group = c.pathParam("group");
            if (state.containsKey(name)) {
                c.fail(409);
            } else {
                addServer(name, group, c.getBodyAsJson(), res -> {
                    if (res.succeeded()) {
                        LOG.log(Level.INFO, "Server \"{0}\" deployed successfully.", name);
                        c.response().end();
                    } else {
                        LOG.log(Level.INFO, "Server \"" + name + "\" failed to deploy.", res.cause());
                        c.fail(500);
                    }
                });
            }
        });
        router.delete().path("/:group/:name").handler(c -> {
            String name = c.pathParam("name");
            String group = c.pathParam("group");
            if (state.containsKey(name) && Objects.equals(state.get(name).getString(Config.CONFIG_LOBBY_GROUP), group)) {
                removeServer(name, res -> {
                    if (res.succeeded()) {
                        LOG.log(Level.INFO, "Server \"{0}\" undeployed successfully.", name);
                        c.response().end();
                    } else {
                        LOG.log(Level.INFO, "Server \"" + name + "\" failed to undeploy.", res.cause());
                        c.fail(500);
                    }
                });
            } else {
                c.fail(404);
            }
        });
        router.get().path("/:group/:name").handler(c -> {
            if (state.containsKey(c.pathParam("name"))) {
                c.response().end(getServer(c.pathParam("name")).toBuffer());
            } else {
                c.fail(404);
            }
        }
        );
        router.get()
                .path("/").handler(c -> {
            if (state.isEmpty()) {
                c.fail(204);
            } else {
                c.response().end(listServers().toBuffer());
            }
        });

        server = vertx.createHttpServer(new HttpServerOptions().setHost(config().getString(Config.CONFIG_MGMT_ADDRESS, "::")).setPort(config().getInteger(Config.CONFIG_MGMT_PORT, 8080)));

        server.requestHandler(router::accept);
        server.listen();
    }

    @Override
    public void stop() throws Exception {
        server.close();
    }

    private void addServer(String name, String serverGroup, JsonObject config, Handler<AsyncResult<String>> handler) {
        int port = new Random().nextInt(60000 - 10000) + 10000;
        JsonObject message = new JsonObject()
                .put(Config.CONFIG_LOBBY_NAME, name)
                .put(Config.CONFIG_LOBBY_GROUP, serverGroup)
                .put(Config.CONFIG_CLIENT_PORT, port)
                .put(Config.CONFIG_LOBBY_CONFIG, config);
        state.put(name, message);
        vertx.eventBus().send(OraLobbyVerticle.EVENTBUS_ADDRESS_ADD, message, res -> {
            if (res.failed()) {
                state.remove(name);
            }
            handler.handle(res.map(Message::body).map(Object::toString));
        });
    }

    private void removeServer(String name, Handler<AsyncResult<String>> handler) {
        vertx.eventBus().send(OraLobbyVerticle.EVENTBUS_ADDRESS_REMOVE, state.get(name), res -> {
            state.remove(name);
            handler.handle(res.map(Message::body).map(Object::toString));
        });
    }

    private JsonObject getServer(String name) {
        return state.get(name).getJsonObject(Config.CONFIG_LOBBY_CONFIG);
    }

    private JsonObject listServers() {
        return state.entrySet().stream().collect(
                JsonObject::new,
                (obj, e) -> obj.put(e.getKey(), e.getValue().getJsonObject(Config.CONFIG_LOBBY_CONFIG)),
                JsonObject::mergeIn);
    }
}
