/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.proxy;

import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.net.NetServer;
import io.vertx.core.net.NetSocket;
import java.util.ArrayList;
import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.packets.DisconnectPacket;
import net.as205315.applications.ora.server.utils.PacketUtils;
import net.as205315.applications.ora.server.utils.SocketAddressParser;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ClientConnectionManager {

    private static final Logger LOG = Logger.getLogger(LobbyServerConnection.class.getName());
    private static final int PROTOCOL_VERSION = 7;

    private final int localPort;
    private final int idBase;
    private final int idLimit;
    private final Consumer<Client> connectHandler;
    private final BiConsumer<Client, Buffer> packetHandler;
    private final Vertx vertx;
    private final NetServer server;
    private final List<Client> clients = new ArrayList<>();

    public ClientConnectionManager(String bindAddress, int localPort, int idBase, int idLimit, Consumer<Client> connectHandler, BiConsumer<Client, Buffer> packetHandler, Vertx vertx) {
        this.localPort = localPort;
        this.idBase = idBase;
        this.idLimit = idLimit;
        this.connectHandler = connectHandler;
        this.packetHandler = packetHandler;
        this.vertx = vertx;

        server = vertx.createNetServer();
        server.connectHandler(this::handleConnection);
        server.listen(localPort, bindAddress, res -> {
            if (res.succeeded()) {
                LOG.log(Level.INFO, "Client server is listening on port {0}.", res.result().actualPort());
            } else {
                LOG.log(Level.SEVERE, "Client server " + localPort + " failed to listen.", res.cause());
            }
        });
    }

    public int getLocalPort() {
        return localPort;
    }

    public Stream<String> getCurrentlyConnectedEndpoints() {
        return clients.stream()
                .map(Client::getSocket)
                .map(NetSocket::remoteAddress)
                .map(SocketAddressParser::toString);
    }

    private void handleConnection(NetSocket socket) {
        int clientId = IntStream.range(idBase, idLimit)
                .filter(i -> clients.stream().map(Client::getId).noneMatch(ci -> ci == i))
                .findFirst()
                .orElseThrow(() -> new IllegalStateException("No free client ID available!"));

        Client client = new Client(clientId, socket);

        socket.closeHandler(v -> {
            clients.remove(client);
            packetHandler.accept(client, new DisconnectPacket(client.getId(), client.getLastFrame()).toBuffer());
        });
        clients.add(client);
        socket.handler(PacketUtils.createReader(Function.identity(), client::receive, true, 0));

        socket.write(Buffer.buffer(8).appendIntLE(PROTOCOL_VERSION).appendIntLE(clientId));
        connectHandler.accept(client);
    }

    public void broadcastDirectExcept(int clientId, Buffer packet) {
        clients.stream()
                .filter(client -> client.isAllowDirect())
                .filter(client -> client.getId() != clientId)
                .forEach(client -> client.send(packet));
    }

    public void broadcast(Buffer packet) {
        clients.stream()
                .forEach(client -> client.send(packet));
    }

    public void broadcastDirect(Buffer packet) {
        clients.stream()
                .filter(client -> client.isAllowDirect())
                .forEach(client -> client.send(packet));
    }

    public void send(int clientId, Buffer packet) {
        clients.stream()
                .filter(client -> client.getId() == clientId)
                .forEach(client -> client.send(packet));
    }

    public void disconnect(int clientId) {
        clients.stream()
                .filter(client -> client.getId() == clientId)
                .forEach(client -> client.disconnect());
    }

    public void setAllowDirect(int clientId, boolean value) {
        clients.stream()
                .filter(client -> client.getId() == clientId)
                .forEach(client -> client.setAllowDirect(value));
    }

    public void close() {
        clients.stream()
                .map(Client::getSocket)
                .forEach(NetSocket::close);

        server.close(v -> {
            LOG.log(Level.INFO, "Client server {0} stopped.", getLocalPort());
        });
    }

    public class Client {

        private final int id;
        private final NetSocket socket;
        private final Consumer<Buffer> sender;
        private int lastFrame = 0;
        private boolean allowDirect = true;

        public Client(int id, NetSocket socket) {
            this.id = id;
            this.socket = socket;
            sender = PacketUtils.createWriter(socket::write, Function.identity(), true, -4);
        }

        public int getId() {
            return id;
        }

        public NetSocket getSocket() {
            return socket;
        }

        public int getLastFrame() {
            return lastFrame;
        }

        public void setLastFrame(int lastFrame) {
            this.lastFrame = lastFrame;
        }

        public boolean isAllowDirect() {
            return allowDirect;
        }

        public void setAllowDirect(boolean allowDirect) {
            this.allowDirect = allowDirect;
        }

        public void receive(Buffer buffer) {
            if (buffer.length() <= 4 || buffer.getUnsignedByte(4) == 0xFF) {
                int frame = buffer.getIntLE(0);
                if (frame > lastFrame) {
                    lastFrame = frame;
                }
            }
            packetHandler.accept(this, Buffer.buffer(4 + buffer.length()).appendUnsignedIntLE(id).appendBuffer(buffer));
        }

        public void send(Buffer buffer) {
            sender.accept(buffer);
        }

        public void disconnect() {
            socket.close();
        }
    }
}
