/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.proxy;

import io.vertx.core.buffer.Buffer;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public abstract class LobbyServerPacket {

    private static final int HEADER_SIZE = 1;

    private final Type type;
    protected final Buffer payloadBuffer;

    private LobbyServerPacket(Type type, Buffer payloadBuffer) {
        this.type = type;
        this.payloadBuffer = payloadBuffer;
    }

    public Type getType() {
        return type;
    }

    public Buffer toBuffer() {
        Buffer buffer = Buffer.buffer(HEADER_SIZE + payloadBuffer.length());

        buffer.appendByte((byte) type.ordinal());
        buffer.appendBuffer(payloadBuffer);

        return buffer;
    }

    public static LobbyServerPacket fromBuffer(Buffer buffer) {
        int type = buffer.getByte(0);
        if (type < 0 || type >= Type.values().length) {
            throw new IllegalArgumentException("Unknown packet type: " + type);
        }
        switch (Type.values()[type]) {
            case STATE:
                return new State(buffer.slice(HEADER_SIZE, buffer.length()));
            case CONNECT_TO:
                return new ConnectTo(buffer.slice(HEADER_SIZE, buffer.length()));
            case START_LOBBY:
                return new StartLobby(buffer.slice(HEADER_SIZE, buffer.length()));
            case STOP_LOBBY:
                return new StopLobby(buffer.slice(HEADER_SIZE, buffer.length()));
            case CLIENT_PACKET:
                return new ClientPacket(buffer.slice(HEADER_SIZE, buffer.length()));
            case CLIENT_PRIVATE_PACKET:
                return new ClientPrivatePacket(buffer.slice(HEADER_SIZE, buffer.length()));
            case CLIENT_FORCE_DISCONNECT:
                return new ClientForceDisconnectPacket(buffer.slice(HEADER_SIZE, buffer.length()));
            case CLIENT_CONNECTED:
                return new ClientConnectedPacket(buffer.slice(HEADER_SIZE, buffer.length()));
            case CLIENT_ALLOW_DIRECT:
                return new ClientAllowDirectPacket(buffer.slice(HEADER_SIZE, buffer.length()));
            default:
                throw new IllegalStateException("Something went wrong when deciding on the packet class for type: " + type);
        }
    }

    public enum Type {
        STATE, CONNECT_TO, START_LOBBY, STOP_LOBBY, CLIENT_PACKET, CLIENT_PRIVATE_PACKET, CLIENT_FORCE_DISCONNECT, CLIENT_CONNECTED, CLIENT_ALLOW_DIRECT;
    }

    public static class State extends LobbyServerPacket {

        private final String ownEndpoint;
        private final List<String> connectedProxies;
        private final List<Integer> runningLobbies;

        private State(Buffer buffer) {
            super(Type.STATE, buffer);

            JsonObject obj = buffer.toJsonObject();
            ownEndpoint = obj.getString("endpoint");
            connectedProxies = obj.getJsonArray("proxies").getList();
            runningLobbies = obj.getJsonArray("lobbies").getList();
        }

        public State(String ownEndpoint, Stream<String> connectedProxies, Stream<Integer> runningLobbies) {
            this(new JsonObject()
                    .put("endpoint", ownEndpoint)
                    .put("proxies", new JsonArray(connectedProxies.collect(Collectors.toList())))
                    .put("lobbies", new JsonArray(runningLobbies.collect(Collectors.toList())))
                    .toBuffer());
        }

        public String getOwnEndpoint() {
            return ownEndpoint;
        }

        public List<String> getConnectedProxies() {
            return connectedProxies;
        }

        public List<Integer> getRunningLobbies() {
            return runningLobbies;
        }
    }

    public static class ConnectTo extends LobbyServerPacket {

        private final String endpoint;

        private ConnectTo(Buffer buffer) {
            super(Type.CONNECT_TO, buffer);

            endpoint = buffer.toString();
        }

        public ConnectTo(String endpoint) {
            this(Buffer.buffer(endpoint));
        }

        public String getEndpoint() {
            return endpoint;
        }
    }

    public static class StartLobby extends LobbyServerPacket {

        private final int port;
        private final int idBase;
        private final int idLimit;

        private StartLobby(Buffer buffer) {
            super(Type.START_LOBBY, buffer);

            port = buffer.getInt(0);
            idBase = buffer.getInt(4);
            idLimit = buffer.getInt(8);
        }

        public StartLobby(int port, int idBase, int idLimit) {
            this(Buffer.buffer(12).appendInt(port).appendInt(idBase).appendInt(idLimit));
        }

        public int getPort() {
            return port;
        }

        public int getIdBase() {
            return idBase;
        }

        public int getIdLimit() {
            return idLimit;
        }
    }

    public static class StopLobby extends LobbyServerPacket {

        private final int port;

        private StopLobby(Buffer buffer) {
            super(Type.STOP_LOBBY, buffer);

            port = buffer.getInt(0);
        }

        public StopLobby(int port) {
            this(Buffer.buffer(4).appendInt(port));
        }

        public int getPort() {
            return port;
        }
    }

    public static class ClientPacket extends LobbyServerPacket {

        private final int lobby;
        private final Buffer packet;

        private ClientPacket(Buffer buffer) {
            super(Type.CLIENT_PACKET, buffer);

            lobby = buffer.getInt(0);
            packet = buffer.slice(4, buffer.length());
        }

        public ClientPacket(int lobby, Buffer packet) {
            this(Buffer.buffer(4 + packet.length()).appendInt(lobby).appendBuffer(packet));
        }

        public int getLobby() {
            return lobby;
        }

        public Buffer getPacket() {
            return packet;
        }
    }

    public static class ClientPrivatePacket extends LobbyServerPacket {

        private final int lobby;
        private final int client;
        private final Buffer packet;

        private ClientPrivatePacket(Buffer buffer) {
            super(Type.CLIENT_PRIVATE_PACKET, buffer);

            lobby = buffer.getInt(0);
            client = buffer.getInt(4);
            packet = buffer.slice(8, buffer.length());
        }

        public ClientPrivatePacket(int lobby, int client, Buffer packet) {
            this(Buffer.buffer(8 + packet.length()).appendInt(lobby).appendInt(client).appendBuffer(packet));
        }

        public int getLobby() {
            return lobby;
        }

        public int getClient() {
            return client;
        }

        public Buffer getPacket() {
            return packet;
        }
    }

    public static class ClientForceDisconnectPacket extends LobbyServerPacket {

        private final int lobby;
        private final int client;

        private ClientForceDisconnectPacket(Buffer buffer) {
            super(Type.CLIENT_FORCE_DISCONNECT, buffer);

            lobby = buffer.getInt(0);
            client = buffer.getInt(4);
        }

        public ClientForceDisconnectPacket(int lobby, int client) {
            this(Buffer.buffer(8).appendInt(lobby).appendInt(client));
        }

        public int getLobby() {
            return lobby;
        }

        public int getClient() {
            return client;
        }
    }

    public static class ClientConnectedPacket extends LobbyServerPacket {

        private final int lobby;
        private final int client;
        private final String remoteAddress;

        private ClientConnectedPacket(Buffer buffer) {
            super(Type.CLIENT_CONNECTED, buffer);

            lobby = buffer.getInt(0);
            client = buffer.getInt(4);
            remoteAddress = buffer.getString(8, buffer.length());
        }

        public ClientConnectedPacket(int lobby, int client, String remoteAddress) {
            this(Buffer.buffer(8).appendInt(lobby).appendInt(client).appendString(remoteAddress));
        }

        public int getLobby() {
            return lobby;
        }

        public int getClient() {
            return client;
        }

        public String getRemoteAddress() {
            return remoteAddress;
        }
    }

    public static class ClientAllowDirectPacket extends LobbyServerPacket {

        private final int lobby;
        private final int client;
        private final boolean allowDirect;

        private ClientAllowDirectPacket(Buffer buffer) {
            super(Type.CLIENT_ALLOW_DIRECT, buffer);

            lobby = buffer.getInt(0);
            client = buffer.getInt(4);
            allowDirect = buffer.getByte(8) != 0;
        }

        public ClientAllowDirectPacket(int lobby, int client, boolean allowDirect) {
            this(Buffer.buffer(8).appendInt(lobby).appendInt(client).appendByte((byte) (allowDirect ? 1 : 0)));
        }

        public int getLobby() {
            return lobby;
        }

        public int getClient() {
            return client;
        }

        public boolean isAllowDirect() {
            return allowDirect;
        }
    }
}
