/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.proxy;

import io.vertx.core.AbstractVerticle;
import io.vertx.core.buffer.Buffer;
import java.util.HashMap;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import net.as205315.applications.ora.server.Config;
import net.as205315.applications.ora.server.utils.SocketAddressParser;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class OraProxyVerticle extends AbstractVerticle {

    private ProxyConnectionManager proxyConnectionManager;
    private LobbyServerConnection lobbyServerConnection;
    private final Map<Integer, ClientConnectionManager> lobbies = new HashMap<>();

    @Override
    public void start() throws Exception {
        proxyConnectionManager = new ProxyConnectionManager(config().getString(Config.CONFIG_PROXY_ADDRESS), config().getInteger(Config.CONFIG_PROXY_PORT), this::handleProxyPacket, vertx);
        lobbyServerConnection = new LobbyServerConnection(config().getString(Config.CONFIG_LOBBY_ADDRESS), config().getInteger(Config.CONFIG_LOBBY_PORT), this::handleLobbyPacket, vertx);

        vertx.setPeriodic(1 * 1000, v -> {
            if (lobbyServerConnection != null && proxyConnectionManager != null) {
                lobbyServerConnection.send(new LobbyServerPacket.State(
                        SocketAddressParser.toString(proxyConnectionManager.getLocalAddress(), proxyConnectionManager.getLocalPort()),
                        proxyConnectionManager.getCurrentlyConnectedEndpoints(),
                        lobbies.keySet().stream()
                ));
            }
        });
    }

    @Override
    public void stop() throws Exception {
        proxyConnectionManager.close();
        lobbyServerConnection.close();
    }

    public void handleLobbyPacket(LobbyServerPacket packet) {
        ClientConnectionManager lobby;
        switch (packet.getType()) {
            case CONNECT_TO:
                proxyConnectionManager.connect((LobbyServerPacket.ConnectTo) packet);
                break;
            case START_LOBBY:
                LobbyServerPacket.StartLobby startLobby = ((LobbyServerPacket.StartLobby) packet);
                if (!lobbies.containsKey(startLobby.getPort())) {
                    lobby = new ClientConnectionManager(config().getString(Config.CONFIG_CLIENT_ADDRESS, "::"), startLobby.getPort(), startLobby.getIdBase(), startLobby.getIdLimit(), handleClientConnected(startLobby.getPort()), handleClientPacket(startLobby.getPort()), vertx);
                    lobbies.put(startLobby.getPort(), lobby);
                }
                break;
            case STOP_LOBBY:
                lobby = lobbies.get(((LobbyServerPacket.StopLobby) packet).getPort());
                if (lobby != null) {
                    lobby.close();
                    lobbies.remove(lobby.getLocalPort());
                }
                break;
            case CLIENT_PACKET:
                lobby = lobbies.get(((LobbyServerPacket.ClientPacket) packet).getLobby());
                if (lobby != null) {
                    lobby.broadcast(((LobbyServerPacket.ClientPacket) packet).getPacket());
                }
                break;
            case CLIENT_PRIVATE_PACKET:
                lobby = lobbies.get(((LobbyServerPacket.ClientPrivatePacket) packet).getLobby());
                if (lobby != null) {
                    lobby.send(((LobbyServerPacket.ClientPrivatePacket) packet).getClient(), ((LobbyServerPacket.ClientPrivatePacket) packet).getPacket());
                }
                break;
            case CLIENT_FORCE_DISCONNECT:
                lobby = lobbies.get(((LobbyServerPacket.ClientForceDisconnectPacket) packet).getLobby());
                if (lobby != null) {
                    lobby.disconnect(((LobbyServerPacket.ClientForceDisconnectPacket) packet).getClient());
                }
                break;
            case CLIENT_ALLOW_DIRECT:
                lobby = lobbies.get(((LobbyServerPacket.ClientAllowDirectPacket) packet).getLobby());
                if (lobby != null) {
                    lobby.setAllowDirect(((LobbyServerPacket.ClientAllowDirectPacket) packet).getClient(), ((LobbyServerPacket.ClientAllowDirectPacket) packet).isAllowDirect());
                }
                break;
            default:
                throw new IllegalArgumentException("Packet type " + packet.getType() + " can't be handled.");
        }
    }

    public void handleProxyPacket(ProxyServerPacket packet) {
        lobbies.get(packet.getLobby()).broadcastDirect(packet.getPayloadBuffer());
    }

    public BiConsumer<ClientConnectionManager.Client, Buffer> handleClientPacket(int lobby) {
        return (client, packet) -> {
            boolean sendDirect;
            if (packet.length() < 9) {
                sendDirect = true;
            } else {
                switch (packet.getUnsignedByte(8)) {
                    case 0x65:
                    case 0xFF:
                        sendDirect = true;
                        break;
                    default:
                        sendDirect = false;
                }
            }

            lobbyServerConnection.send(new LobbyServerPacket.ClientPacket(lobby, packet));

            if (sendDirect && client.isAllowDirect()) {
                proxyConnectionManager.broadcast(new ProxyServerPacket(lobby, packet));
                lobbies.get(lobby).broadcastDirectExcept(packet.getIntLE(0), packet);
            }
        };
    }

    public Consumer<ClientConnectionManager.Client> handleClientConnected(int lobby) {
        return client -> {
            lobbyServerConnection.send(new LobbyServerPacket.ClientConnectedPacket(lobby, client.getId(), SocketAddressParser.toString(client.getSocket().remoteAddress())));
        };
    }
}
