/*
 * Copyright (C) 2018 Jean-Rémy Buchs <jrb0001@692b8c32.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package net.as205315.applications.ora.server.proxy;

import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.net.NetServer;
import io.vertx.core.net.NetSocket;
import java.net.InetSocketAddress;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import net.as205315.applications.ora.server.utils.PacketUtils;
import net.as205315.applications.ora.server.utils.SocketAddressParser;

/**
 *
 * @author Jean-Rémy Buchs <jrb0001@692b8c32.de>
 */
public class ProxyConnectionManager {

    private static final Logger LOG = Logger.getLogger(LobbyServerConnection.class.getName());

    private final String localAddress;
    private final Consumer<ProxyServerPacket> packetHandler;
    private final Vertx vertx;
    private final Set<NetSocket> sockets = new HashSet<>();
    private final NetServer server;

    public ProxyConnectionManager(String localAddress, int localPort, Consumer<ProxyServerPacket> packetHandler, Vertx vertx) {
        this.localAddress = localAddress;
        this.packetHandler = packetHandler;
        this.vertx = vertx;

        server = vertx.createNetServer();
        server.connectHandler(this::handleConnection);
        server.listen(localPort, localAddress, res -> {
            if (res.succeeded()) {
                LOG.log(Level.INFO, "Proxy server is listening on port {0}.", res.result().actualPort());
            } else {
                LOG.log(Level.SEVERE, "Proxy server failed to listen.", res.cause());
            }
        });
    }

    public String getLocalAddress() {
        return localAddress;
    }

    public int getLocalPort() {
        return server.actualPort();
    }

    public Stream<String> getCurrentlyConnectedEndpoints() {
        return sockets.stream()
                .map(NetSocket::remoteAddress)
                .map(SocketAddressParser::toString);
    }

    public void connect(LobbyServerPacket.ConnectTo packet) {
        if (getCurrentlyConnectedEndpoints().noneMatch(packet.getEndpoint()::equals)) {
            InetSocketAddress endpoint = SocketAddressParser.fromString(packet.getEndpoint());
            LOG.log(Level.INFO, "Connecting to proxy server: {0}:{1}.", new Object[]{endpoint.getAddress().getHostAddress(), endpoint.getPort()});
            vertx.createNetClient().connect(endpoint.getPort(), endpoint.getAddress().getHostAddress(), res -> {
                if (res.succeeded()) {
                    LOG.log(Level.INFO, "Connected to proxy server: {0}:{1}.", new Object[]{endpoint.getAddress().getHostAddress(), endpoint.getPort()});
                    NetSocket socket = res.result();
                    socket.closeHandler(v -> {
                        sockets.remove(socket);
                    });

                    handleConnection(socket);
                    sockets.add(socket);
                } else {
                    LOG.log(Level.SEVERE, "Failed to connect to proxy server " + endpoint + ".", res.cause());
                }
            });
        }
    }

    private void handleConnection(NetSocket socket) {
        socket.handler(PacketUtils.createReader(ProxyServerPacket::fromBuffer, packetHandler::accept, false, 0));
    }

    public void broadcast(ProxyServerPacket packet) {
        sockets.stream()
                .map(socket -> (Consumer<Buffer>) socket::write)
                .map(PacketUtils.createWriterMapper(ProxyServerPacket::toBuffer, false, 0))
                .forEach(consumer -> consumer.accept(packet));
    }

    public void close() {
        sockets.stream()
                .forEach(NetSocket::close);

        server.close(v -> {
            LOG.log(Level.INFO, "Proxy server stopped.");
        });
    }
}
